/**
 *  @brief                                                                  
 *  kripke.c - Kripke*SAT 0.1                                               
 *                                                                          
 *  @author Valentin Montmirail (montmirail@cril.fr)                        
 *                                                                          
 */

#include "kripke.h"

/**
 * @brief This function will return the number of edges used in the Kripke model.
 *
 * @param kripke the Kripke structure which store the edges.
 *
 * @return the number of edges which are connecting worlds together.
 */
inline int getNbEdges(st_kripke* kripke) {

  /* at the beginning, we don't know how many edges there is, so 0. */
  int result = 0;

  int i = 0;
  int j = 0;
  
  /* We will check the edge FROM each world w_i */
  for( ; i < kripke->num_worlds; i++) {

      /* We will check the edge TO each world w_j */
      for(j = 0; j < kripke->num_worlds; j++) {
      
            /* 
             * Then we increase the number of edges. Or by +0 or by +1. 
             * (Can't be something else) 
             */
            result += (kripke->edges[i][j]);            
      }
  }

  /* And we return at the end how many edges there is in the Kripke structure */
  return result;
}

/**
 * @brief This function will tell us if our parameter is a negate value or not.
 *
 * It will return 1 if vertex is a NOT(phi), and 0 otherwise.
 *
 * @param vertex the sub-formula where we want to check the negativity
 * @return 1 if vertex is a NOT(phi), and 0 otherwise.
 */
inline int isNegative(vertex_t* vertex) {

  return V_op_code(vertex) == not_code;
}

/**
 * @brief This function will tell us if our parameter is 2 block of litterals linked by an 'and' or an 'or'.
 *
 * It will return 1 if vertex is a 'phi and psi' or 'phi or psi', and 0 otherwise.
 *
 * @param vertex the sub-formula where we want to check the operator
 * @return 1 if vertex is a 'phi and psi' or 'phi or psi', and 0 otherwise.
 */
inline int isBinaryOperator(vertex_t* vertex) {
  return  ((V_op_code(vertex) == and_code) || (V_op_code(vertex) == or_code));
}

/**
 * @brief This function will tell us if the vertex can be use in the Tableau method or not.
 *
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param vertex the branch that we are testing.
 * 
 * @return 1 if the vertex can be use, 0 otherwise.
 */
inline int accepted_value(dag_t* dag_ptr, vertex_t* vertex,vertex_t* branch) {

  lsGen         list_generator;
  lsGeneric     t_lsGeneric;
  stGeneric     k_stGeneric, v_stGeneric;
  st_generator *t_st_generator;

  int branchIsNegative = 0;

  /* if branchIsNegative is equals to 1, the branch is negative, 0 otherwise. */
  if(branch != NULL) {
    branchIsNegative = isNegative(branch);
  } 

  st_foreach_item(dag_ptr->mv_table_ptr, t_st_generator, &k_stGeneric, &v_stGeneric) {
      
      lsForEachItem(*UNCAST_VALUE(v_stGeneric), list_generator, t_lsGeneric) {

          if(!branchIsNegative && Are_dag_vtx_equal(UNCAST_VERTEX_ITEM(t_lsGeneric),vertex)) {
            return 1;
          }



          if(branchIsNegative && Are_dag_vtx_equal(UNCAST_VERTEX_ITEM(t_lsGeneric),vertex)) {

            return 2;
          }
      }
  }


  /* If here reach here, then the branch is not accepted. */

  return 0;
} 

/**
 * @brief This function will tell us if the vertex can be use in the Tableau method or not.
 *
 * @param a_vertex_ptr the branch that we are testing.
 * @param firstTime 1 if it's the first time we call this function for this vertex, 0 otherwise.
 * 
 * @return 1 if the vertex can be use, 0 otherwise.
 */
int authorised_value(vertex_t *a_vertex_ptr,int firstTime) {
  
  int       result = 0;
  
  switch (V_op_code(a_vertex_ptr)) {

    case and_code:   
      result = 1; 
      break; 

    case or_code:           
      result = 1; 
      break; 

    case not_code:
      if(firstTime) {
        return ;
      }
      result = 1; 
      break;    
      
    case atom_code:
      result = 1; 
      break;
      
    case top_code:  
      result = 1; 
      break; 

    case bot_code:     
      result = 1;
      break; 
  }

  return result; 
} 

/**
 * @brief This function will initialize a Kripke structure with the good number of variables
 *
 * But the number of worlds may fluctuate (depends of the solving method by the Tableau).
 *
 * @param nbVar the number of propositionals variables.
 * @param nbWorld the original number of worlds (basically the number of modal variables).
 *
 * @return the initialized Kripke structure. 
 */
st_kripke* createStrucKripke(int nbVar, int nbWorld) {

    int i;
    int j;

    /* We need to allocate the place for one Kripke structure. */
    st_kripke* kripke = (st_kripke*)malloc(sizeof(st_kripke)*1);

    kripke->num_variables = nbVar;
    kripke->num_worlds = nbWorld;
    
    /* edges is a matrix [nbWorlds][nbWorlds] so we need to initialize it as an array of arrays */
    kripke->edges = (int**)malloc(sizeof(int*)*nbWorld);
    
    /* For each world w_i */
    for(i=0; i < nbWorld; i++) {
      
      /* We initialize the array edges[i] with the size nbWorld (in order to create a matrix) */
      kripke->edges[i] = (int*)malloc(sizeof(int)*nbWorld);  

      /* For each world w_j */
      for(j=0; j < nbWorld; j++) {

        kripke->edges[i][j] = 0;
      }

    }

    /* Then we will need to store each assignations for each world */

    /* So we allocate the place to store an array of integers for every world. */
    kripke->assignations = (int**)malloc(sizeof(int*)*nbWorld);

    /* For each world w_i, we allocate an array with the size nbVar to create a matrix [nbWorld][nbVar] */
    for(i = 0; i < nbWorld; i++) {

      kripke->assignations[i] = (int*)malloc(sizeof(int)*nbVar);

      /* Then we initialize each case of theses array with a 0. */
      for(j = 0; j < nbVar; j++) {
        kripke->assignations[i][j] = 0;
      }
    }

    /* By default, we know that we have one 0, w_0 which is always existing. */
    kripke->num_worlds = 1;

    /* Then we return the perfectly initialized structure. */
    return kripke;
}

/**
 * @brief This function will initialized the Table that we use as an Oracle to perform the Tableau method.
 * 
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param image_ptr the pointer to the image structure which is the link between our modal variable and their CNF translation.
 * @param unknown the number of unknown variables (the last index of variable - (number of propositional variables + modal variables)
 */
void initializeAuthorizedValuesTable(dag_t* dag_ptr, image_t *image_ptr, int unknown) {

  stGeneric     k_stGeneric, v_stGeneric;
  lsGen         list_generator;
  lsGeneric     t_lsGeneric;
  st_generator *t_st_generator;
  vertex_t* item_vertex_ptr;
  int i;
  int authorized = 0;

  /** If we want to use the Tableau method on Phi, Phi should be allowed ;). */
  Add_to_table(dag_ptr->mv_table_ptr,UNCAST_VERTEX_ITEM(dag_ptr->root_vertex_ptr));

  /* For every propositional variable, we will add them to our authorized values table. */
  For_each_prop_xref_item(dag_ptr, i, item_vertex_ptr) {

    if (item_vertex_ptr != NIL(vertex_t)) {
      
      /* 
       * We are adding this value from here if and only if this value is returned positive by the SAT solver.
       * This comes from the fact that dag->ptr is storing only positive values.
       * So if the SAT solver return s -1 ..., no need to add [C0] from here,
       * We will need to add (NOT C0) later. 
       */
      if (Value(image_ptr)[i+1] == TT) {
        Add_to_table(dag_ptr->mv_table_ptr,UNCAST_VERTEX_ITEM(item_vertex_ptr));
      } 

      if (Value(image_ptr)[i+1] == FF) {
        Add_to_table(dag_ptr->mv_table_ptr,UNCAST_VERTEX_ITEM( Negate_vertex(dag_ptr,item_vertex_ptr)));
      } 

    }
  }

  For_each_modal_xref_item(dag_ptr, i, item_vertex_ptr) {

    /* 
     * If one modal variable represents in fact, an 'or' between other variables.
     * it should be allowed to use in the tableau method, to avoid missing assignations. 
     */
    if (Value(image_ptr)[i+1] == TT) {
        Add_to_table(dag_ptr->mv_table_ptr,item_vertex_ptr);
    }

    if (Value(image_ptr)[i+1] == FF) {
      Add_to_table(dag_ptr->mv_table_ptr,Negate_vertex(dag_ptr,item_vertex_ptr));
    } 

  }

  st_foreach_item(dag_ptr->uni_st_table_ptr, t_st_generator, &k_stGeneric, &v_stGeneric) {
    lsForEachItem(*UNCAST_VALUE(v_stGeneric), list_generator, t_lsGeneric) {

      item_vertex_ptr = UNCAST_VERTEX_ITEM(t_lsGeneric);

      /* 
       * If the variable is a litteral (P ; ~P) 
       * (even if it's a modal litteral) we are able to use them in the Tableau
       */
      if(Is_literal_vertex(item_vertex_ptr)) {

        /* We add this variable into our "mv Table", the table of authorized value in the Tableau. */
        Add_to_table(dag_ptr->mv_table_ptr,item_vertex_ptr);
      }

      /* If there is still something to add : */
      if(unknown != 0) {

        /* If the value is authorized. */
        authorized = authorised_value(item_vertex_ptr,1);

        if(authorized == 1) {
            
            /* Then we add it and we decrement the number of unkown. */
            Add_to_table(dag_ptr->mv_table_ptr,item_vertex_ptr);
            unknown--;
        }
      } 
    } /* end of the lsForEachItem */
  } /* end of the st_foreach_item */
}

/**
 * @brief This function is a debug function, 
 *
 * it will display the Hash table of all the authorized values for the Tableau method.
 * 
 * @param dag_ptr the pointer to our DAG Structure.
 */
void displayAuthorizedValues(dag_t *dag_ptr) {

  lsGen         list_generator;
  lsGeneric     t_lsGeneric;
  stGeneric     k_stGeneric, v_stGeneric;
  st_generator *t_st_generator;

  st_foreach_item(dag_ptr->mv_table_ptr, t_st_generator, &k_stGeneric, &v_stGeneric) {
      
      lsForEachItem(*UNCAST_VALUE(v_stGeneric), list_generator, t_lsGeneric) {

          /* We print on 5 characters, the Hash key of our vertex */
          printf("c %5d - ", UNCAST_KEY(k_stGeneric));

          /* And then we display our vertex itself. */
          Print_vertex_rec(UNCAST_VERTEX_ITEM(t_lsGeneric));

          printf("\n");
      }
  }
}

/**
 * 
 * @brief This function a debugging function, it will display:
 *
 * - which are the propositionals variables.
 * - which are the modals variabls.
 * - What are the allowed values for the Tableau method.
 * 
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param image_ptr the pointer to the image structure which is the link between our modal variable and their CNF translation.
 * 
 */
void print_details(dag_t *dag_ptr,image_t *image_ptr) {
  
  int i;
  int nbKnownVariables = 0;
  int unknown = 0;
  vertex_t* item_vertex_ptr;
    
    /* For every propositinal variable, we will display its name C_i */
    For_each_prop_xref_item(dag_ptr, i, item_vertex_ptr) {

        /* Only if this variable is different than NIL (which normally will not happen) */
        if (item_vertex_ptr != NIL(vertex_t)) {

  		   /* printf("c  C%d\n",i); */
          ++nbKnownVariables;
        }
    }


    /* Then we display all the modal variables */
    For_each_modal_xref_item(dag_ptr, i, item_vertex_ptr) {
      
      /* First the name of the modal variable C_i */
      
      /* printf("c C%d : ",i); */

      /* Then the real value ([]C_0 for example) */
  	  /* Print_vertex_rec(V_ini_vertex_ptr(item_vertex_ptr)); */

      ++nbKnownVariables;
      
      /* printf("\n"); */
    }  
          
  unknown = (Max_atom(image_ptr) - nbKnownVariables);

  /* We first initialized our Authorized values. */
  initializeAuthorizedValuesTable(dag_ptr,image_ptr, unknown);
} 

/**  
 * @brief This function will display all the knowledges that we've got.
 * 
 * Which are : 
 *   - the SAT assignemnt returned by the SAT solver.
 *   - The list of all the propositionals variables.
 *   - The list of all the modals variables.
 *   - The list of all the authorized values for the Tableau method.
 *
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param image_ptr the pointer to the image structure which is the link between our modal variable and their CNF translation.
 * 
 */
void displayInputKnowledges(dag_t *dag_ptr, image_t *image_ptr) {

  /*
  printf("c sat ");
  print_model (image_ptr);
  */

  /*
  printf("c\nc original formula ");
  Print_vertex_rec(UNCAST_VERTEX_ITEM(dag_ptr->root_vertex_ptr));
  printf("\n");
  */

  print_details(dag_ptr,image_ptr);

}

/**
 * @brief This function will give us the Kripke model in a Kripke-CNF format.
 *
 * It will also perform the extraction from the SAT solution.
 *
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param image_ptr the pointer to the image structure which is the link between our modal variable and their CNF translation.
 * 
 */
void displayKripkeModel(dag_t *dag_ptr, image_t *image_ptr) {

  int i = 0;
  int j = 0;
  int k = 0;
  int nbEdges = 0;

  /* 
   * We first create the Kripke structure.
   * - Lpa() represents the last index of propositionals variables, 
   *         it's starting at 0, so +1 is the number of propositionals variables.
   * - Lma() represents the last index of modals variables so (Lma - number of propositional variables)
   *         is giving us the number of modal variables.
   */
  st_kripke * kripke = createStrucKripke(Lpa(dag_ptr) +1,Lma(dag_ptr) - Lpa(dag_ptr) +1);

  /* We display some input knowledges, to know from where we start. */
  displayInputKnowledges(dag_ptr,image_ptr);  

  /* Then we start the extraction from the SAT solution by the Tableau method. */
  extraction(dag_ptr,kripke);

  /* After extraction, we display the authorized values that we used */
  /* displayAuthorizedValues(dag_ptr); */

  /* Every line starting with a 'c' is a comment. */
  printf("c #var #worlds #relations #edges\n");

  if(solvercalls > 1) solvercalls--;

  printf("d SOLVERCALLS %d\n",solvercalls);

  nbEdges = getNbEdges(kripke);

  printf("v %d %d 1 %d\n",kripke->num_variables-1,kripke->num_worlds,nbEdges);

  for(i=0; i < kripke->num_worlds; i++) {

    j = 0;

    printf("v ");

    /* We are printing each assignations for every world (until we find a 0, the end of the assignation) */
    for(j = 0; j < kripke->num_variables || kripke->assignations[i][j] == 0 ; j++) { 
      
      if(kripke->assignations[i][j] != 0 && abs(kripke->assignations[i][j]) <= kripke->num_variables) {
        int val = kripke->assignations[i][j];
        if(val > 0) {
          val--;
        } else {
          val++;
        }
        printf("%d ",val); 
      }

    }
    printf("0\n");
  }

  /* 
   * After we finished to display all the assignations for every world,
   * We need to display the relations between each world.
   */
  for(i = 0 ; i < kripke->num_worlds; i++) {
      for(j = 0; j < kripke->num_worlds; j++) {
          
          /* If there is a relation, we display this relation. */
          if(kripke->edges[i][j]) {

            printf("v r1 w%d w%d\n",i,j);
          }
      }
  }
  
}

/**
 * @brief This function is starting the Tableau method from the initial formula
 *
 * It also says that our original world is w_0.
 * 
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 */
void extraction(dag_t* dag_ptr,st_kripke * kripke) {

    /* We always start from w_0, even in a simple SAT problem. */
    int  world = 0;

    vertex_t* item_vertex_ptr;
    int i = 0;

    /* Then we start the recursivity in the Tableau. */
    extraction_rec(dag_ptr->root_vertex_ptr,dag_ptr,kripke,world);
}

/**
 * @brief This function will set the value assignations[world][j] at the good value.
 * 
 * if item_vertex_ptr = branch, then assignations[world][j] = (i+1)
 *
 * if item_vertex_ptr = not(branch), then assignations[world][j] = (-1)*(i+1) 
 *
 * @param world the current world that we are getting the assignments for each variables.
 * @param i the value of the propositional variable that we need to assign in the world.
 * @param j the j-th assigned value for the current world.
 * @param item_vertex_ptr the value equals to 'branch' but item_vertex_ptr is always the "true" value.
 * @param branch the branch of the tree that we are using to know if the variable should be set at false or true.
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 */
void assignation(int world,int i, int j, vertex_t *item_vertex_ptr, vertex_t* branch,st_kripke* kripke,dag_t* dag_ptr) {

    int k = 0;
    int alreadyDone = 0;
    int twoNegatives = 0;

    if(branch != item_vertex_ptr) {
      twoNegatives = (isNegative(branch)) && (isNegative(item_vertex_ptr));
    }  

    /* We will first check if we didn't assigned somehow the value 'i+1' to avoid duplicate. */
    for(k = 0; k < kripke->num_variables; k++) {
      if( kripke->assignations[world][k] == (i+1) || kripke->assignations[world][k] == (-1)*(i+1)) {
        alreadyDone = 1;
      }
    }

    /* 
     * If it's the first time that we are putting the value, 
     * then we are putting assignations[world][j] = (i+1) 
     */
    if(!alreadyDone) {

      if(accepted_value(dag_ptr,item_vertex_ptr,branch)) {

        if(isNegative(item_vertex_ptr)) {
          kripke->assignations[world][j] = (-1)*(i+1);
        } else {
            kripke->assignations[world][j] = (i+1);
        }
      }
    }      
}

/**
 * @brief This function will set the value assignations[world][j] at the good value.
 * 
 * if item_vertex_ptr = branch, then assignations[world][j] = (i+1)
 *
 * if item_vertex_ptr = not(branch), then assignations[world][j] = (-1)*(i+1) 
 * 
 * if item_vertex_ptr = branch(modal-variable), then we call addElement with the good content of this modal variable.
 *
 * @warning The difference with assignation, is that here the branch contains a modal variable.
 *
 * @param world the current world that we are getting the assignments for each variables.
 * @param i the value of the propositional variable that we need to assign in the world.
 * @param j the j-th assigned value for the current world.
 * @param item_vertex_ptr the value equals to 'branch' but item_vertex_ptr is always the "true" value.
 * @param branch the branch of the tree that we are using to know if the variable should be set at false or true.
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 */
void assignationModal(int world,int i, int j, vertex_t *item_vertex_ptr, vertex_t* branch,st_kripke* kripke,dag_t* dag_ptr) {

  lsList        sons_vertices_lsList;
  st_generator* t_st_generator;
  vertex_t* modal_vertex1,* modal_vertex2,*modal_vertex_ptr;
  vertex_t      *son_vertex,*prop_vertex_ptr;
  lsGen         list_generator;
  lsGeneric     t_lsGeneric;

  int prop = 0;
  int k = 0;
  int alreadyDone = 0;
  int stillModal = 1;
  int accepted = 0;  

  if(isNegative(branch)) {
      item_vertex_ptr = Negate_vertex(dag_ptr,item_vertex_ptr);
  }


  /** For every modal variable C_i */
  For_each_modal_xref_item(dag_ptr, i, modal_vertex_ptr) {
    
    /** We are checking if branch = modal_vertex or branch = not(modal_vertex) */
    if(Are_dag_vtx_equal(UNCAST_VERTEX_ITEM(modal_vertex_ptr),branch) || Are_dag_vtx_compl_eq(UNCAST_VERTEX_ITEM(modal_vertex_ptr),branch) ) {
      
      /* We are getting the 'inside' of our modal variable (if C_i = [ ]C_j, we are getting C_j) */
      son_vertex = Get_vertex_son(V_ini_vertex_ptr(modal_vertex_ptr)); 

      if(isNegative(branch)) {
        son_vertex = Negate_vertex(dag_ptr,son_vertex);
      }

      /* For each propositional variable C_prop */
      For_each_prop_xref_item(dag_ptr, prop, prop_vertex_ptr) {

        /* If the content of our "modal variable" = prop_vertex or not(prop_vertex) */
        if(Are_dag_vtx_equal(UNCAST_VERTEX_ITEM(prop_vertex_ptr),son_vertex) || 
           Are_dag_vtx_compl_eq(UNCAST_VERTEX_ITEM(prop_vertex_ptr),son_vertex)
          ) {

          /* 
           * If we find this variable inside the list of propositionals variables, then we are not modal anymore
           * But it can happen that C_3 = []C2 and C2 = []~C0, and then when we get C2, we don't find it in the 
           * list of propositional variable, so we have to recursively call this function in order to get the C0.
           */

           stillModal = 0;

           assignation(world,prop,j,son_vertex, branch,kripke,dag_ptr);  
        }
      }
    }
  }

  /*
   * It can happen that C_3 = []C2 and C2 = []~C0, and then when we get C2, we don't find it in the 
   * list of propositional variables, 
   * so we have to recursively call addElement in order to get C0 (to assign in the good world).
   */
  if(stillModal) {

    /* Then again, we have to check if the branch is with a NOT or if it's not the case. */
    alreadyDone = 0;
    
    /* This is the content of our branch, which is unfortunately, still modal */
    /* Print_vertex_rec(son_vertex); */

    /* We will go through every modal variables */
    For_each_modal_xref_item(dag_ptr, i, modal_vertex1) {

      /* If branch is NOT() */
      if(Are_dag_vtx_compl_eq(UNCAST_VERTEX_ITEM(modal_vertex1), branch)) { 

        /* We will go again through every modal variable */
        For_each_modal_xref_item(dag_ptr, i, modal_vertex2) {

          /* And if son_vertex is NOT(), then the 2 are cancelling each other. */
          if(Are_dag_vtx_compl_eq(UNCAST_VERTEX_ITEM(modal_vertex1), son_vertex)) { 

            /* Then we did the assignation already */
            alreadyDone = 1;

            accepted = accepted_value(dag_ptr,Get_vertex_son(son_vertex),branch);
            if(accepted) {

              /* We remove the negation of son_vertex */
              Add_to_table(dag_ptr->mv_table_ptr,Get_vertex_son(son_vertex));
            }

              extraction_rec(Get_vertex_son(son_vertex),dag_ptr,kripke,world);      
          }
        }
      }
    }

    /* If it's was 0 NOT, or only 1, no need to change son_vertex before calling addElement. */
    if(!alreadyDone) {

      if(authorised_value(son_vertex,1)) {
          Add_to_table(dag_ptr->mv_table_ptr,son_vertex);
      }


      /* Once may notice that because we perform the rule of a [], we increase the index of the world. */
      extraction_rec(son_vertex,dag_ptr,kripke,world);
      
    }
  }   
}

/**
 * @brief This function will make the call to assignation or assignationModal at the good indexes.
 *
 * It's also here that the relation between worlds are affected (everytime a modality is detected).
 * 
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param branch the branch of the tree that we are using to know if the variable should be set at false or true.
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 * @param world the index of our current world (0 at the beginning, after depends on the recursives calls).
 */
void addElement(dag_t* dag_ptr, vertex_t* branch,st_kripke * kripke, int world) {
          
    vertex_t* item_vertex_ptr;
    int i = 0;
    int j = 0;

    /* For each propositional variable stored in the DAG structure */
    For_each_prop_xref_item(dag_ptr, i, item_vertex_ptr) {

      j = 0;
      if (item_vertex_ptr != NIL(vertex_t)) {

        /* If branch = item_vertex_ptr or branch = not(item_vertex_ptr) */
        if(Are_dag_vtx_equal(UNCAST_VERTEX_ITEM(item_vertex_ptr),branch) || Are_dag_vtx_compl_eq(UNCAST_VERTEX_ITEM(item_vertex_ptr),branch) ) {
          
          /* We will look for the first place free to assigned the value of the variable. */
          while(kripke->assignations[world][j] != 0) {
            j++;
          }

          /* 
           * Then we call assignation, 
           * which is the function that will put into the Kripke structure, 
           * the good assignation for (i+1) 
           */
           assignation(world,i,j,branch,branch,kripke,dag_ptr);
        }
      }
    }    

    /* For each modal variable stored in the DAG structure */
    For_each_modal_xref_item(dag_ptr, i, item_vertex_ptr) {

      j = 0;
      if (item_vertex_ptr != NIL(vertex_t)) {

        /* If branch = item_vertex_ptr or branch = not(item_vertex_ptr) */
        if(Are_dag_vtx_equal(UNCAST_VERTEX_ITEM(item_vertex_ptr),branch) || Are_dag_vtx_compl_eq(UNCAST_VERTEX_ITEM(item_vertex_ptr),branch) ) {

          /* We find a modality, so we need to indicate the relation between the current world and the world of the modality */
          kripke->edges[kripke->num_worlds-1][world] = 1;

          /* We will have, a new world globally in the Kripke model. */
          if(world >= kripke->num_worlds) {
            kripke->num_worlds++;
          }

          /* We will look for the first place free to assigned the value of the variable. */
          while(kripke->assignations[world][j] != 0) {
            j++;
          }

          /* 
           * Then we call assignationModal, 
           * which is the function that will put into the Kripke structure, 
           * the good assignation for (i+1) in the good world.
           */
           assignationModal(world,i,j,item_vertex_ptr,branch,kripke,dag_ptr);
        }
      }
    }    
}

/**
 * @brief 
 * This function will extract recursively all the branch of our formula
 * 
 * Let's say our formula is : `phi = ((A or B) and ((C and D) and (E or F)))`

 * It's will be call as follow :
 *   - extraction_rec((A or B))
 *        - extraction_rec (A)
 *        - extraction_rec (B)
 *   - extraction_rec((C and D))
 *        - extraction_rec (C)
 *        - extraction_rec (D)
 *   - extraction_rec((E or F))
 *        - extraction_rec (E)
 *        - extraction_rec (F)
 *
 * Basically it's a depth-first going through a tree.
 * 
 * @warning It's here that we are using our forbidden list, to know if the element is allowed to be assigned or not.
 *
 * @param branch the branch of the tree that we are using to know if the variable should be set at false or true.
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 * @param world the index of our current world (0 at the beginning, after depends on the recursives calls).
 */
void extraction_rec(vertex_t* branch,dag_t* dag_ptr,st_kripke * kripke, int world) {

  lsList        sons_vertices_lsList, out_edges_lsList;
  vertex_t      *son_vertex_ptr,*father_vertex_ptr, *new_box_vertex_ptr;
  lsGen         list_generator;
  stGeneric     k_stGeneric, v_stGeneric;
  lsStatus      list_status;
  st_generator *t_st_generator;
  lsGeneric     t_lsGeneric;
  lsList        fathers;
  vertex_t      *item_vertex_ptr;
  int authorized = 0;
  int negation = 0;

  if(branch == NIL(vertex_t)) {
    return ;
  }

  if(isNegative(branch) && isNegative(Get_vertex_son(branch))) {
    branch = Get_vertex_son(Get_vertex_son(branch));
  }

  authorized = accepted_value(dag_ptr,branch,NULL);

  /* If and only if, branch is included into mv (or if we reach a leaf */
  if(authorized) {
 
    /* If our branch is a litteral (can be a modal litteral also) */
    if(Is_literal_vertex(branch)) {

      int i = 0;
      /* For each modal variable C_i */
      For_each_modal_xref_item(dag_ptr, i, item_vertex_ptr) {
        
        if (item_vertex_ptr != NIL(vertex_t)) {
          /* If branch = item_vertex_ptr or branch = not(item_vertex_ptr) */
          if(Are_dag_vtx_compl_eq(UNCAST_VERTEX_ITEM(item_vertex_ptr),branch) || Are_dag_vtx_equal(UNCAST_VERTEX_ITEM(item_vertex_ptr),branch)) {
            
            /* We increase the number of world because it's a <> */
            world = world+1;
            break;
          }
        }
      }

      /* Then we call addElement and we return to break the recursivity, we reach a leaf */
      addElement(dag_ptr,branch,kripke,world);  

      return ;
    }

    if(isNegative(branch)) {
        negation = 1;
        branch = Get_vertex_son(branch);
    }

    /* If we are here, we are not a litteral, so the branch has 'sons' */
    sons_vertices_lsList = Get_vertex_sons(branch);  

    /* We will go through every son, and call recursively this function. */
    lsForEachItem(sons_vertices_lsList, list_generator, t_lsGeneric) {

        son_vertex_ptr = UNCAST_VERTEX_ITEM(t_lsGeneric);

        if(negation) {
          son_vertex_ptr = Negate_vertex(dag_ptr,son_vertex_ptr);
        }

        extraction_rec(son_vertex_ptr,dag_ptr,kripke,world);  
    }
    negation = 0;
  } 
}