/**
 *  @brief                                                                  
 *  kripke.h - Kripke*SAT 0.1                                               
 *                                                                          
 *  @author Valentin Montmirail (montmirail@cril.fr)                        
 *                                                                          
 */

/* Standard modules */
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>

/* *SAT modules */
#include "set.h"

/** @brief We rename the structure, to avoid typing struct everywhere... */
typedef struct st_kripke st_kripke;

/**
 * @brief This structure is how we store the Kripke model. 
 *
 * We so have informations about the number of propositional variables, the number of worlds in the Kripke model.
 *
 * But also the assignations for each world and how worls are connected to each other.
 */
struct st_kripke {
    
    /** @brief How many propositional variables there are in the problem. */
    int num_variables;

    /** @brief How many worlds we use to generate the Kripke model. */
    int num_worlds;

    /** @brief What are the edges from world w_i to world w_j (store as a Matrix [nbWorlds][nbWorlds]). */
    int** edges;

    /** @brief What are the assignations of propositionals variables for each world (store as a Matrix [nbWorlds][nbVariables]). */
    int** assignations;
};

/**
 * @brief This function will return the number of edges used in the Kripke model.
 *
 * @param kripke the Kripke structure which store the edges.
 *
 * @return the number of edges which are connecting worlds together.
 */
inline int getNbEdges(st_kripke* kripke);

/**
 * @brief This function will tell us if our parameter is a negate value or not.
 *
 * It will return 1 if vertex is a NOT(phi), and 0 otherwise.
 *
 * @param vertex the sub-formula where we want to check the negativity
 * @return 1 if vertex is a NOT(phi), and 0 otherwise.
 */
inline int isNegative(vertex_t* vertex);

/**
 * @brief This function will tell us if our parameter is 2 block of litterals linked by an 'and' or an 'or'.
 *
 * It will return 1 if vertex is a 'phi and psi' or 'phi or psi', and 0 otherwise.
 *
 * @param vertex the sub-formula where we want to check the operator
 * @return 1 if vertex is a 'phi and psi' or 'phi or psi', and 0 otherwise.
 */
inline int isBinaryOperator(vertex_t* vertex);

/**
 * @brief This function will tell us if the vertex can be use in the Tableau method or not.
 *
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param vertex the branch that we are testing.
 * 
 * @return 1 if the vertex can be use, 0 otherwise.
 */
inline int accepted_value(dag_t* dag_ptr, vertex_t* vertex,vertex_t* branch);

/**
 * @brief This function will tell us if the vertex can be use in the Tableau method or not.
 *
 * @param a_vertex_ptr the branch that we are testing.
 * @param firstTime 1 if it's the first time we call this function for this vertex, 0 otherwise.
 * 
 * @return 1 if the vertex can be use, 0 otherwise.
 */
int authorised_value_for_mv(vertex_t *a_vertex_ptr,int firstTime);

/**
 * @brief This function will initialize a Kripke structure with the good number of variables
 *
 * But the number of worlds may fluctuate (depends of the solving method by the Tableau).
 *
 * @param nbVar the number of propositionals variables.
 * @param nbWorld the original number of worlds (basically the number of modal variables).
 *
 * @return the initialized Kripke structure. 
 */
st_kripke* createStrucKripke(int nbVar, int nbWorld);

/**
 * @brief This function will initialized the Table that we use as an Oracle to perform the Tableau method.
 * 
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param image_ptr the pointer to the image structure which is the link between our modal variable and their CNF translation.
 * @param unknown the number of unknown variables (the last index of variable - (number of propositional variables + modal variables)
 */
void initializeAuthorizedValuesTable(dag_t* dag_ptr, image_t *image_ptr, int unknown);

/**
 * @brief This function is a debug function, 
 *
 * it will display the Hash table of all the authorized values for the Tableau method.
 * 
 * @param dag_ptr the pointer to our DAG Structure.
 */
void displayAuthorizedValues(dag_t *dag_ptr);

/**
 * 
 * @brief This function a debugging function, it will display:
 *
 * - which are the propositionals variables.
 * - which are the modals variabls.
 * - What are the allowed values for the Tableau method.
 * 
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param image_ptr the pointer to the image structure which is the link between our modal variable and their CNF translation.
 * 
 */
void print_details(dag_t *dag_ptr,image_t *image_ptr);

/**  
 * @brief This function will display all the knowledges that we've got.
 * 
 * Which are : 
 *   - the SAT assignemnt returned by the SAT solver.
 *   - The list of all the propositionals variables.
 *   - The list of all the modals variables.
 *   - The list of all the authorized values for the Tableau method.
 *
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param image_ptr the pointer to the image structure which is the link between our modal variable and their CNF translation.
 * 
 */
void displayInputKnowledges(dag_t *dag_ptr, image_t *image_ptr);

/**
 * @brief This function will give us the Kripke model in a Kripke-CNF format.
 *
 * It will also perform the extraction from the SAT solution.
 *
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param image_ptr the pointer to the image structure which is the link between our modal variable and their CNF translation.
 * 
 */
void displayKripkeModel(dag_t *dag_ptr, image_t *image_ptr);

/**
 * @brief This function is starting the Tableau method from the initial formula
 *
 * It also says that our original world is w_0.
 * 
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 */
void extraction(dag_t* dag_ptr,st_kripke * kripke);

/**
 * @brief This function will set the value assignations[world][j] at the good value.
 * 
 * if item_vertex_ptr = branch, then assignations[world][j] = (i+1)
 *
 * if item_vertex_ptr = not(branch), then assignations[world][j] = (-1)*(i+1) 
 *
 * @param world the current world that we are getting the assignments for each variables.
 * @param i the value of the propositional variable that we need to assign in the world.
 * @param j the j-th assigned value for the current world.
 * @param item_vertex_ptr the value equals to 'branch' but item_vertex_ptr is always the "true" value.
 * @param branch the branch of the tree that we are using to know if the variable should be set at false or true.
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 */
void assignation(int world,int i, int j, vertex_t *item_vertex_ptr, vertex_t* branch,st_kripke* kripke,dag_t* dag_ptr);

/**
 * @brief This function will set the value assignations[world][j] at the good value.
 * 
 * if item_vertex_ptr = branch, then assignations[world][j] = (i+1)
 *
 * if item_vertex_ptr = not(branch), then assignations[world][j] = (-1)*(i+1) 
 * 
 * if item_vertex_ptr = branch(modal-variable), then we call addElement with the good content of this modal variable.
 *
 * @warning The difference with assignation, is that here the branch contains a modal variable.
 *
 * @param world the current world that we are getting the assignments for each variables.
 * @param i the value of the propositional variable that we need to assign in the world.
 * @param j the j-th assigned value for the current world.
 * @param item_vertex_ptr the value equals to 'branch' but item_vertex_ptr is always the "true" value.
 * @param branch the branch of the tree that we are using to know if the variable should be set at false or true.
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 */
void assignationModal(int world,int i, int j, vertex_t *item_vertex_ptr, vertex_t* branch,st_kripke* kripke,dag_t* dag_ptr);

/**
 * @brief This function will make the call to assignation or assignationModal at the good indexes.
 *
 * It's also here that the relation between worlds are affected (everytime a modality is detected).
 * 
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param branch the branch of the tree that we are using to know if the variable should be set at false or true.
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 * @param world the index of our current world (0 at the beginning, after depends on the recursives calls).
 */
void addElement(dag_t* dag_ptr, vertex_t* branch,st_kripke * kripke, int world);

/**
 * @brief 
 * This function will extract recursively all the branch of our formula
 * 
 * Let's say our formula is : `phi = ((A or B) and ((C and D) and (E or F)))`

 * It's will be call as follow :
 *   - extraction_rec((A or B))
 *        - extraction_rec (A)
 *        - extraction_rec (B)
 *   - extraction_rec((C and D))
 *        - extraction_rec (C)
 *        - extraction_rec (D)
 *   - extraction_rec((E or F))
 *        - extraction_rec (E)
 *        - extraction_rec (F)
 *
 * Basically it's a depth-first going through a tree.
 * 
 * @warning It's here that we are using our forbidden list, to know if the element is allowed to be assigned or not.
 *
 * @param branch the branch of the tree that we are using to know if the variable should be set at false or true.
 * @param dag_ptr the pointer to our DAG structure (which store all the variables).
 * @param kripke the pointer to our Kripke structure, (where results will be store).
 * @param world the index of our current world (0 at the beginning, after depends on the recursives calls).
 */
void extraction_rec(vertex_t* branch,dag_t* dag_ptr,st_kripke * kripke, int world);

